import io
import json
import zipfile
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Hashable, Iterable, Iterator
from functools import cached_property
from pathlib import Path
from typing import Generic, cast

import questionary
import typer
from cognite.client import data_modeling as dm
from cognite.client.data_classes import (
    DataSetList,
    ExtractionPipelineList,
    Group,
    GroupList,
    TransformationList,
    TransformationNotificationList,
    TransformationScheduleList,
    filters,
)
from cognite.client.data_classes._base import (
    CogniteResourceList,
)
from cognite.client.data_classes.agents import (
    AgentList,
)
from cognite.client.data_classes.data_modeling import DataModelId
from cognite.client.data_classes.documents import SourceFileProperty
from cognite.client.data_classes.extractionpipelines import ExtractionPipelineConfigList
from cognite.client.data_classes.functions import (
    Function,
    FunctionList,
    FunctionSchedulesList,
)
from cognite.client.data_classes.workflows import (
    Workflow,
    WorkflowList,
    WorkflowTriggerList,
    WorkflowVersion,
    WorkflowVersionId,
    WorkflowVersionList,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils import ms_to_datetime
from questionary import Choice
from rich import print
from rich.console import Console
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import LocationFilterList
from cognite_toolkit._cdf_tk.client.data_classes.search_config import SearchConfigList
from cognite_toolkit._cdf_tk.client.data_classes.search_config import ViewId as SearchConfigViewId
from cognite_toolkit._cdf_tk.client.data_classes.streamlit_ import Streamlit, StreamlitList
from cognite_toolkit._cdf_tk.cruds import (
    AgentCRUD,
    ContainerCRUD,
    DataModelCRUD,
    DataSetsCRUD,
    ExtractionPipelineConfigCRUD,
    ExtractionPipelineCRUD,
    FunctionCRUD,
    FunctionScheduleCRUD,
    GroupCRUD,
    LocationFilterCRUD,
    NodeCRUD,
    ResourceCRUD,
    SearchConfigCRUD,
    SpaceCRUD,
    StreamlitCRUD,
    TransformationCRUD,
    TransformationNotificationCRUD,
    TransformationScheduleCRUD,
    ViewCRUD,
    WorkflowCRUD,
    WorkflowTriggerCRUD,
    WorkflowVersionCRUD,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import T_ID
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceRetrievalError,
    ToolkitMissingResourceError,
    ToolkitResourceMissingError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import FileExistsWarning, HighSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree, safe_write, sanitize_filename, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.interactive_select import DataModelingSelect

from ._base import ToolkitCommand

_INTERACTIVE_SELECT_HELPER_TEXT = " Use arrow keys to navigate and space key to select. Press enter to confirm."


class ResourceFinder(Iterable, ABC, Generic[T_ID]):
    def __init__(self, client: ToolkitClient, identifier: T_ID | None = None):
        self.client = client
        self.identifier = identifier

    def _selected(self) -> T_ID:
        return self.identifier or self._interactive_select()

    @abstractmethod
    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        raise NotImplementedError

    @abstractmethod
    def _interactive_select(self) -> T_ID:
        raise NotImplementedError

    # Can be implemented in subclasses
    def update(self, resources: CogniteResourceList) -> None: ...


class DataModelFinder(ResourceFinder[DataModelId]):
    def __init__(self, client: ToolkitClient, identifier: DataModelId | None = None, include_global: bool = False):
        super().__init__(client, identifier)
        self._include_global = include_global
        self.data_model: dm.DataModel[dm.ViewId] | None = None
        self.view_ids: set[dm.ViewId] = set()
        self.container_ids: set[dm.ContainerId] = set()
        self.space_ids: set[str] = set()

    def _interactive_select(self) -> DataModelId:
        data_model_ids = self.client.data_modeling.data_models.list(
            all_versions=False, limit=-1, include_global=False
        ).as_ids()
        available_spaces = sorted({model.space for model in data_model_ids})
        if not available_spaces:
            raise ToolkitMissingResourceError("No data models found")
        if len(available_spaces) == 1:
            selected_space = available_spaces[0]
        else:
            selected_space = questionary.select("In which space is your data model located?", available_spaces).ask()
        data_model_ids = sorted(
            [model for model in data_model_ids if model.space == selected_space], key=lambda model: model.as_tuple()
        )

        selected_data_model: DataModelId = questionary.select(
            "Which data model would you like to dump?",
            [
                Choice(f"{model_id!r}", value=model_id)
                for model_id in sorted(data_model_ids, key=lambda model: model.as_tuple())
            ],
        ).ask()

        retrieved_models = self.client.data_modeling.data_models.retrieve(
            (selected_data_model.space, selected_data_model.external_id), inline_views=False
        )
        if not retrieved_models:
            # This happens if the data model is removed after the list call above.
            raise ToolkitMissingResourceError(f"Data model {selected_data_model} not found")
        if len(retrieved_models) == 1:
            self.data_model = retrieved_models[0]
            return selected_data_model
        models_by_version = {model.version: model for model in retrieved_models if model.version is not None}
        if len(models_by_version) == 1:
            self.data_model = retrieved_models[0]
            return selected_data_model
        if not questionary.confirm(
            f"Would you like to select a different version than {selected_data_model.version} of the data model",
            default=False,
        ).ask():
            self.data_model = models_by_version[cast(str, selected_data_model.version)]
            return selected_data_model

        selected_model = questionary.select(
            "Which version would you like to dump?",
            [
                Choice(f"{version} ({len(model.views)} views)", value=version)
                for version, model in models_by_version.items()
            ],
        ).ask()
        self.data_model = models_by_version[selected_model]
        return self.data_model.as_id()

    def update(self, resources: CogniteResourceList) -> None:
        if isinstance(resources, dm.DataModelList):
            self.view_ids |= {
                view.as_id() if isinstance(view, dm.View) else view for item in resources for view in item.views
            }
        elif isinstance(resources, dm.ViewList):
            self.container_ids |= resources.referenced_containers()
        elif isinstance(resources, dm.SpaceList):
            return
        self.space_ids |= {item.space for item in resources}

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        model_loader = DataModelCRUD.create_loader(self.client)
        if self.data_model:
            is_global_model = self.data_model.is_global
            yield [], dm.DataModelList([self.data_model]), model_loader, None
        else:
            model_list = model_loader.retrieve([self.identifier])
            if not model_list:
                raise ToolkitResourceMissingError(f"Data model {self.identifier} not found", str(self.identifier))
            is_global_model = model_list[0].is_global
            yield [], model_list, model_loader, None
        if self._include_global or is_global_model:
            yield list(self.view_ids), None, ViewCRUD.create_loader(self.client), "views"
            yield list(self.container_ids), None, ContainerCRUD.create_loader(self.client), "containers"
            yield list(self.space_ids), None, SpaceCRUD.create_loader(self.client), None
        else:
            view_loader = ViewCRUD(self.client, None, None, topological_sort_implements=True)
            views = dm.ViewList([view for view in view_loader.retrieve(list(self.view_ids)) if not view.is_global])
            yield [], views, view_loader, "views"
            container_loader = ContainerCRUD.create_loader(self.client)
            containers = dm.ContainerList(
                [
                    container
                    for container in container_loader.retrieve(list(self.container_ids))
                    if not container.is_global
                ]
            )
            yield [], containers, container_loader, "containers"

            space_loader = SpaceCRUD.create_loader(self.client)
            spaces = dm.SpaceList(
                [space for space in space_loader.retrieve(list(self.space_ids)) if not space.is_global]
            )
            yield [], spaces, space_loader, None


class WorkflowFinder(ResourceFinder[WorkflowVersionId]):
    def __init__(self, client: ToolkitClient, identifier: WorkflowVersionId | None = None):
        super().__init__(client, identifier)
        self._workflow: Workflow | None = None
        self._workflow_version: WorkflowVersion | None = None

    def _interactive_select(self) -> WorkflowVersionId:
        workflows = self.client.workflows.list(limit=-1)
        if not workflows:
            raise ToolkitMissingResourceError("No workflows found")
        selected_workflow_id: str = questionary.select(
            "Which workflow would you like to dump?",
            [Choice(workflow_id, value=workflow_id) for workflow_id in workflows.as_external_ids()],
        ).ask()
        for workflow in workflows:
            if workflow.external_id == selected_workflow_id:
                self._workflow = workflow
                break

        versions = self.client.workflows.versions.list(selected_workflow_id, limit=-1)
        if len(versions) == 0:
            raise ToolkitMissingResourceError(f"No versions found for workflow {selected_workflow_id}")
        if len(versions) == 1:
            self._workflow_version = versions[0]
            return self._workflow_version.as_id()

        selected_version: WorkflowVersionId = questionary.select(
            "Which version would you like to dump?",
            [Choice(f"{version!r}", value=version) for version in versions.as_ids()],
        ).ask()
        for version in versions:
            if version.version == selected_version.version:
                self._workflow_version = version
                break
        return selected_version

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        if self._workflow:
            yield [], WorkflowList([self._workflow]), WorkflowCRUD.create_loader(self.client), None
        else:
            yield [self.identifier.workflow_external_id], None, WorkflowCRUD.create_loader(self.client), None
        if self._workflow_version:
            yield (
                [],
                WorkflowVersionList([self._workflow_version]),
                WorkflowVersionCRUD.create_loader(self.client),
                None,
            )
        else:
            yield [self.identifier], None, WorkflowVersionCRUD.create_loader(self.client), None
        trigger_loader = WorkflowTriggerCRUD.create_loader(self.client)
        trigger_list = WorkflowTriggerList(
            list(trigger_loader.iterate(parent_ids=[self.identifier.workflow_external_id]))
        )
        yield [], trigger_list, trigger_loader, None


class TransformationFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.transformations: TransformationList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.transformations = self.client.transformations.list(limit=-1)
        if self.transformations and not any(transformation.external_id for transformation in self.transformations):
            raise ToolkitValueError(
                "ExternalID is required for dumping transformations. "
                f"Found {len(self.transformations)} transformations with only internal IDs."
            )
        elif not self.transformations:
            raise ToolkitMissingResourceError("No transformations found")

        choices = [
            Choice(f"{transformation.name} ({transformation.external_id})", value=transformation.external_id)
            for transformation in sorted(self.transformations, key=lambda t: t.name or "")
            if transformation.external_id
        ]

        selected_transformation_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which transformation(s) would you like to dump?",
            choices=choices,
        ).ask()
        if not selected_transformation_ids:
            raise ToolkitValueError(f"No transformations selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_transformation_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        if self.transformations:
            yield (
                [],
                TransformationList([t for t in self.transformations if t.external_id in self.identifier]),
                TransformationCRUD.create_loader(self.client),
                None,
            )
        else:
            yield list(self.identifier), None, TransformationCRUD.create_loader(self.client), None

        schedule_loader = TransformationScheduleCRUD.create_loader(self.client)
        schedule_list = TransformationScheduleList(list(schedule_loader.iterate(parent_ids=list(self.identifier))))
        yield [], schedule_list, schedule_loader, None
        notification_loader = TransformationNotificationCRUD.create_loader(self.client)
        notification_list = TransformationNotificationList(
            list(notification_loader.iterate(parent_ids=list(self.identifier)))
        )
        yield [], notification_list, notification_loader, None


class GroupFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.groups: list[Group] | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        groups = self.client.iam.groups.list(all=True)
        if not groups:
            raise ToolkitMissingResourceError("No groups found")
        groups_by_name: dict[str, list[Group]] = defaultdict(list)
        for group in groups:
            groups_by_name[group.name].append(group)
        selected_groups: list[list[Group]] | None = questionary.checkbox(
            "Which group(s) would you like to dump?",
            choices=[
                Choice(f"{group_name} ({len(group_list)} group{'s' if len(group_list) > 1 else ''})", value=group_list)
                for group_name, group_list in sorted(groups_by_name.items())
            ],
        ).ask()
        if not selected_groups:
            raise ToolkitValueError(f"No group selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        self.groups = [group for group_list in selected_groups for group in group_list]
        return tuple(group_list[0].name for group_list in selected_groups)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        if self.groups:
            yield [], GroupList(self.groups), GroupCRUD.create_loader(self.client), None
        else:
            yield list(self.identifier), None, GroupCRUD.create_loader(self.client), None


class AgentFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.agents: AgentList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.agents = self.client.agents.list()
        if not self.agents:
            raise ToolkitMissingResourceError("No agents found")

        choices = [
            Choice(f"{agent.name} ({agent.external_id}) with {len(agent.tools)} tools", value=agent.external_id)
            for agent in sorted(self.agents, key=lambda a: a.name or a.external_id)
            if agent.external_id
        ]

        selected_agent_ids: list[str] | None = questionary.checkbox(
            "Which agent(s) would you like to dump?",
            choices=choices,
        ).ask()
        if not selected_agent_ids:
            raise ToolkitValueError(f"No agents selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_agent_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = AgentCRUD.create_loader(self.client)
        if self.agents:
            yield (
                [],
                AgentList([agent for agent in self.agents if agent.external_id in self.identifier]),
                loader,
                None,
            )
        else:
            yield list(self.identifier), None, loader, None


class NodeFinder(ResourceFinder[dm.ViewId]):
    def __init__(self, client: ToolkitClient, identifier: dm.ViewId | None = None):
        super().__init__(client, identifier)
        self.is_interactive = False

    def _interactive_select(self) -> dm.ViewId:
        self.is_interactive = True
        spaces = self.client.data_modeling.spaces.list(limit=-1)
        if not spaces:
            raise ToolkitMissingResourceError("No spaces found")
        selected_space: str = questionary.select(
            "In which space is your node property view located?", [space.space for space in spaces]
        ).ask()

        views = self.client.data_modeling.views.list(space=selected_space, limit=-1, all_versions=False)
        if not views:
            raise ToolkitMissingResourceError(f"No views found in {selected_space}")
        if len(views) == 1:
            return views[0].as_id()
        selected_view_id: dm.ViewId = questionary.select(
            "Which node property view would you like to dump?",
            [Choice(repr(view), value=view) for view in views.as_ids()],
        ).ask()
        return selected_view_id

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = NodeCRUD(self.client, None, None, self.identifier)
        if self.is_interactive:
            count = self.client.data_modeling.instances.aggregate(
                self.identifier, dm.aggregations.Count("externalId"), instance_type="node"
            ).value
            if count == 0 or count is None:
                raise ToolkitMissingResourceError(f"No nodes found in {self.identifier}")
            elif count > 50:
                if not questionary.confirm(
                    f"Are you sure you want to dump {count} nodes? This may take a while.",
                    default=False,
                ).ask():
                    typer.Exit(0)
        nodes = dm.NodeList[dm.Node](list(loader.iterate()))
        yield [], nodes, loader, None


class LocationFilterFinder(ResourceFinder[tuple[str, ...]]):
    @cached_property
    def all_filters(self) -> LocationFilterList:
        return self.client.search.locations.list()

    def _interactive_select(self) -> tuple[str, ...]:
        filters = self.all_filters
        if not filters:
            raise ToolkitMissingResourceError("No filters found")
        id_by_display_name = {f"{filter.name} ({filter.external_id})": filter.external_id for filter in filters}
        selected_filter_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which filters would you like to dump?",
            choices=[Choice(name, value=id_) for name, id_ in id_by_display_name.items()],
        ).ask()
        if not selected_filter_ids:
            raise ToolkitValueError(f"No filters selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_filter_ids)

    def _get_filters(self, identifiers: tuple[str, ...]) -> LocationFilterList:
        if not identifiers:
            return self.all_filters
        filters = [f for f in self.all_filters if f.external_id in identifiers]
        if not filters:
            raise ToolkitResourceMissingError(f"Location filters {identifiers} not found", str(identifiers))
        return LocationFilterList(filters)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self.identifier or self._interactive_select()
        filters = self._get_filters(self.identifier)
        yield [], filters, LocationFilterCRUD.create_loader(self.client), None


class ExtractionPipelineFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.extraction_pipelines: ExtractionPipelineList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.extraction_pipelines = self.client.extraction_pipelines.list(limit=-1)
        if not self.extraction_pipelines:
            raise ToolkitMissingResourceError("No extraction pipelines found")
        choices = [
            Choice(f"{pipeline.name} ({pipeline.external_id})", value=pipeline.external_id)
            for pipeline in sorted(self.extraction_pipelines, key=lambda p: p.name or "")
            if pipeline.external_id
        ]
        selected_pipeline_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which extraction pipeline(s) would you like to dump?",
            choices=choices,
        ).ask()
        if not selected_pipeline_ids:
            raise ToolkitValueError(f"No extraction pipelines selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_pipeline_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        pipeline_loader = ExtractionPipelineCRUD.create_loader(self.client)
        if self.extraction_pipelines:
            selected_pipelines = ExtractionPipelineList(
                [p for p in self.extraction_pipelines if p.external_id in self.identifier]
            )
            yield [], selected_pipelines, pipeline_loader, None
        else:
            yield list(self.identifier), None, pipeline_loader, None
        config_loader = ExtractionPipelineConfigCRUD.create_loader(self.client)
        configs = ExtractionPipelineConfigList(list(config_loader.iterate(parent_ids=list(self.identifier))))
        yield [], configs, config_loader, None


class DataSetFinder(ResourceFinder[tuple[str, ...]]):
    """Finds data sets to dump."""

    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.datasets: DataSetList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.datasets = self.client.data_sets.list(limit=-1)
        if not self.datasets:
            raise ToolkitMissingResourceError("No datasets found")
        choices = [
            Choice(f"{dataset.name} ({dataset.external_id})", value=dataset.external_id)
            for dataset in sorted(self.datasets, key=lambda d: d.name or "")
            if dataset.external_id
        ]
        selected_dataset_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which dataset(s) would you like to dump?",
            choices=choices,
        ).ask()
        if not selected_dataset_ids:
            raise ToolkitValueError(f"No datasets selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_dataset_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = DataSetsCRUD.create_loader(self.client)
        if self.datasets:
            yield (
                [],
                DataSetList([d for d in self.datasets if d.external_id in set(self.identifier)]),
                loader,
                None,
            )
        else:
            yield list(self.identifier), None, loader, None


class FunctionFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.functions: FunctionList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.functions = self.client.functions.list(limit=-1)
        if not self.functions:
            raise ToolkitMissingResourceError("No functions found")
        choices = [
            Choice(f"{function.name} ({function.external_id})", value=function.external_id)
            for function in sorted(self.functions, key=lambda f: f.name or "")
            if function.name and function.external_id
        ]
        selected_function_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which function(s) would you like to dump?",
            choices=choices,
        ).ask()
        if not selected_function_ids:
            raise ToolkitValueError(f"No functions selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_function_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = FunctionCRUD.create_loader(self.client)
        if self.functions:
            selected_functions = FunctionList([f for f in self.functions if f.external_id in self.identifier])
            yield [], selected_functions, loader, None
        else:
            yield list(self.identifier), None, loader, None

        schedule_loader = FunctionScheduleCRUD.create_loader(self.client)
        schedules = schedule_loader.iterate(parent_ids=list(self.identifier))
        yield [], FunctionSchedulesList(list(schedules)), schedule_loader, None

    def dump_function_code(self, function: Function, folder: Path) -> None:
        try:
            zip_bytes = self.client.files.download_bytes(id=function.file_id)
        except CogniteAPIError as e:
            if e.code == 400 and "File ids not found" in e.message:
                HighSeverityWarning(
                    f"The function {function.external_id!r} does not have code to dump. It is not available in CDF."
                ).print_warning()
                return
            raise
        try:
            top_level = f"{sanitize_filename(function.external_id or 'unknown_external_id')}/"
            with zipfile.ZipFile(io.BytesIO(zip_bytes)) as zf:
                if all(name.startswith(top_level) for name in zf.namelist()):
                    zf.extractall(folder)
                else:
                    zf.extractall(folder / top_level)
        except zipfile.BadZipFile as e:
            HighSeverityWarning(
                f"The function {function.external_id!r} has a corrupted code zip file. Unable to extract code: {e!s}"
            ).print_warning()


class StreamlitFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.apps: StreamlitList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        """Interactively select one or more Streamlit apps to dump."""
        result = self.client.documents.aggregate_unique_values(
            SourceFileProperty.metadata_key("creator"),
            filter=filters.Equals(SourceFileProperty.directory, "/streamlit-apps/"),
        )
        if not result:
            raise ToolkitMissingResourceError("No Streamlit apps found")

        selected_creator = questionary.select(
            "Who is the creator of the Streamlit app you would like to dump? [name (app count)]",
            choices=[
                Choice(f"{item.value} ({item.count})", value=item.value)
                for item in sorted(result, key=lambda r: (r.count, str(r.value) or ""))
            ],
        ).ask()
        files = self.client.files.list(
            limit=-1, directory_prefix="/streamlit-apps/", metadata={"creator": str(selected_creator)}
        )
        self.apps = StreamlitList([Streamlit.from_file(file) for file in files if file.name and file.external_id])
        if missing := [file for file in files if not file.external_id or file.name]:
            MediumSeverityWarning(
                f"{len(missing)} file(s) in /streamlit-apps/ are missing "
                f"either name or external ID and will be skipped. File IDs: {humanize_collection([file.id for file in missing])}",
            ).print_warning()
        selected_ids: list[str] | None = questionary.checkbox(
            message="Which Streamlit app(s) would you like to dump?",
            choices=[
                Choice(
                    title=f"{app.name} ({app.creator} - {ms_to_datetime(app.last_updated_time)})", value=app.external_id
                )
                for app in sorted(self.apps, key=lambda a: a.name)
            ],
        ).ask()
        if not selected_ids:
            raise ToolkitValueError(f"No Streamlit app selected for dumping.{_INTERACTIVE_SELECT_HELPER_TEXT}")
        return tuple(selected_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        identifier = self.identifier or self._interactive_select()
        loader = StreamlitCRUD.create_loader(self.client)
        # If the user used interactive select, we have already downloaded the streamlit apps,
        # Thus, we do not need to download them again. If not pass the identifier and let the main logic
        # take care of the download.
        if self.apps:
            yield [], StreamlitList([app for app in self.apps if app.external_id in identifier]), loader, None
        else:
            yield list(identifier), None, loader, None

    def dump_code(self, app: Streamlit, folder: Path, console: Console | None = None) -> None:
        """Dump the code of a Streamlit app to the specified folder.

        The code is extracted from the JSON content of the app file in CDF.

        Args:
            app (Streamlit): The Streamlit app whose code is to be dumped.
            folder (Path): The directory where the app code will be saved.
            console (Console | None): Optional Rich console for printing warnings.
        """
        try:
            content = self.client.files.download_bytes(external_id=app.external_id)
        except CogniteAPIError as e:
            if e.code == 400 and e.missing:
                HighSeverityWarning(
                    f"The source code for {app.external_id!r} could not be retrieved from CDF."
                ).print_warning(console=console)
                return
            raise

        try:
            json_content = json.loads(content)
        except json.JSONDecodeError as e:
            HighSeverityWarning(
                f"The JSON content for the Streamlit app {app.external_id!r} is corrupt and could not be extracted. "
                f"Download file with the same external id manually to remediate. {e!s}"
            ).print_warning(console=console)
            return

        app_folder = sanitize_filename(app.external_id)
        app_path = folder / app_folder
        app_path.mkdir(exist_ok=True)
        if isinstance(json_content.get("requirements"), list):
            requirements_txt = app_path / "requirements.txt"
            requirements = json_content["requirements"]
            if not requirements:
                HighSeverityWarning(
                    f"The Streamlit app {app.external_id!r} has a requirements.txt file with no content. Skipping..."
                ).print_warning()
            else:
                requirements_txt.write_text("\n".join(requirements), encoding="utf-8")
        files = json_content.get("files", {})
        if not isinstance(files, dict) or (not files):
            HighSeverityWarning(
                f"The Streamlit app {app.external_id!r} does not have any files to dump. It is likely corrupted."
            ).print_warning(console=console)
            return
        created_files: set[str] = set()
        for relative_filepath, content in files.items():
            filepath = app_path / relative_filepath
            filepath.parent.mkdir(parents=True, exist_ok=True)
            file_content = content.get("content", {}).get("text", "")
            if not isinstance(file_content, str):
                HighSeverityWarning(
                    f"The Streamlit app {app.external_id!r} has a file {relative_filepath} with invalid content. Skipping..."
                ).print_warning(console=console)
                continue
            safe_write(filepath, file_content, encoding="utf-8")
            created_files.add(relative_filepath)

        entry_point = json_content.get("entrypoint")
        if entry_point and entry_point not in created_files:
            HighSeverityWarning(
                f"The Streamlit app {app.external_id!r} has an entry point {entry_point} that was not found in the files. "
                "The app may be corrupted."
            ).print_warning(console=console)


class SpaceFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)

    def _interactive_select(self) -> tuple[str, ...]:
        # Using new interactive select component for selecting instance spaces
        # This will raise a ToolkitValueError if no instance spaces are selected and this is handled in the caller.
        data_modeling_select = DataModelingSelect(self.client, "dump")
        selected_spaces = data_modeling_select.select_instance_space(
            multiselect=True, message="Which instance space(s) would you like to dump?"
        )
        return tuple(selected_spaces)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = SpaceCRUD.create_loader(self.client)
        yield list(self.identifier), None, loader, None


class SearchConfigFinder(ResourceFinder[tuple[SearchConfigViewId, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[SearchConfigViewId, ...] | None = None):
        super().__init__(client, identifier)
        self.search_configs: SearchConfigList | None = None

    def _interactive_select(self) -> tuple[SearchConfigViewId, ...]:
        self.search_configs = self.client.search.configurations.list()
        if not self.search_configs:
            raise ToolkitMissingResourceError("No search configurations found!")
        choices = [
            Choice(f"{config.view.external_id} {config.view.space}", value=config.view)
            for config in self.search_configs
        ]
        selected_view_ids: list[SearchConfigViewId] | None = questionary.checkbox(
            "For which view would you like to dump the search configuration?",
            choices=choices,
        ).ask()
        if not selected_view_ids:
            raise ToolkitValueError("No view selected for dumping the search configuration.")
        return tuple(selected_view_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceCRUD, None | str]]:
        self.identifier = self._selected()
        loader = SearchConfigCRUD.create_loader(self.client)
        if self.search_configs:
            yield [], SearchConfigList([sc for sc in self.search_configs if sc.view in self.identifier]), loader, None
        else:
            yield list(self.identifier), None, loader, None


class DumpResourceCommand(ToolkitCommand):
    def dump_to_yamls(
        self,
        finder: ResourceFinder,
        output_dir: Path,
        clean: bool,
        verbose: bool,
    ) -> None:
        is_populated = output_dir.exists() and any(output_dir.iterdir())
        if is_populated and clean:
            safe_rmtree(output_dir)
            output_dir.mkdir()
            self.console(f"Cleaned existing output directory {output_dir!s}.")
        elif is_populated:
            self.warn(MediumSeverityWarning("Output directory is not empty. Use --clean to remove existing files."))
        elif not output_dir.exists():
            output_dir.mkdir(exist_ok=True)

        dumped_ids: list[Hashable] = []
        for identifiers, resources, loader, subfolder in finder:
            if not identifiers and not resources:
                # No resources to dump
                continue
            if resources is None:
                try:
                    resources = loader.retrieve(identifiers)
                except CogniteAPIError as e:
                    raise ResourceRetrievalError(f"Failed to retrieve {humanize_collection(identifiers)}: {e!s}") from e
                if len(resources) == 0:
                    raise ToolkitResourceMissingError(
                        f"Resource(s) {humanize_collection(identifiers)} not found", str(identifiers)
                    )

            finder.update(resources)
            resource_folder = output_dir / loader.folder_name
            if subfolder:
                resource_folder = resource_folder / subfolder
            resource_folder.mkdir(exist_ok=True, parents=True)
            for resource in resources:
                resource_id = loader.get_id(resource)
                name = loader.as_str(resource_id)
                base_filepath = resource_folder / f"{name}.{loader.kind}.yaml"
                if base_filepath.exists():
                    self.warn(FileExistsWarning(base_filepath, "Skipping... Use --clean to remove existing files."))
                    continue
                dumped = loader.dump_resource(resource)
                for filepath, subpart in loader.split_resource(base_filepath, dumped):
                    content = subpart if isinstance(subpart, str) else yaml_safe_dump(subpart)
                    safe_write(filepath, content, encoding="utf-8")
                    if verbose:
                        self.console(f"Dumped {loader.kind} {name} to {filepath!s}")
                if isinstance(finder, FunctionFinder) and isinstance(resource, Function):
                    finder.dump_function_code(resource, resource_folder)
                if isinstance(finder, StreamlitFinder) and isinstance(resource, Streamlit):
                    finder.dump_code(resource, resource_folder)
                dumped_ids.append(resource_id)
        print(Panel(f"Dumped {humanize_collection(dumped_ids)}", title="Success", style="green", expand=False))



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
import shutil
import time
import warnings
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from time import sleep

import questionary
from cognite.client.data_classes.capabilities import (
    AssetsAcl,
    Capability,
    FunctionsAcl,
    GroupsAcl,
    ProjectsAcl,
    RelationshipsAcl,
    SessionsAcl,
)
from cognite.client.data_classes.iam import Group, GroupList, GroupWrite, TokenInspection
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.panel import Panel
from rich.prompt import Prompt
from rich.table import Table

from cognite_toolkit._cdf_tk import cruds
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.constants import (
    HINT_LEAD_TEXT,
    TOOLKIT_DEMO_GROUP_NAME,
    TOOLKIT_SERVICE_PRINCIPAL_GROUP_NAME,
)
from cognite_toolkit._cdf_tk.exceptions import (
    AuthenticationError,
    AuthorizationError,
    ResourceCreationError,
    ResourceDeleteError,
    ToolkitMissingValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    LowSeverityWarning,
    MediumSeverityWarning,
    MissingCapabilityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables, prompt_user_environment_variables

from ._base import ToolkitCommand


@dataclass
class VerifyAuthResult:
    toolkit_group_id: int | None = None
    function_status: str | None = None


class AuthCommand(ToolkitCommand):
    def init(self, no_verify: bool = False, dry_run: bool = False) -> None:
        env_vars: EnvironmentVariables | None = None
        try:
            env_vars = EnvironmentVariables.create_from_environment()
        except ToolkitMissingValueError:
            ...

        ask_user = True
        if env_vars and not env_vars.get_missing_vars():
            print("Auth variables are already set.")
            ask_user = questionary.confirm("Do you want to reconfigure the auth variables?", default=False).ask()

        if ask_user or not env_vars:
            env_vars = prompt_user_environment_variables(env_vars)
            self._store_dotenv(env_vars)

        client = env_vars.get_client()
        try:
            client.iam.token.inspect()
        except CogniteAPIError as e:
            raise AuthenticationError(f"Unable to verify the credentials.\n{e}")

        print("[green]The credentials are valid.[/green]")
        if no_verify:
            return
        print(
            Panel(
                "Running verification, 'cdf auth verify'...",
                title="",
                expand=False,
            )
        )
        self.verify(client, dry_run)

    def _store_dotenv(self, env_vars: EnvironmentVariables) -> None:
        new_env_file = env_vars.create_dotenv_file()
        if Path(".env").exists():
            existing = Path(".env").read_text(encoding="utf-8")
            if existing == new_env_file:
                print("Identical '.env' file already exist.")
                return None
            self.warn(MediumSeverityWarning("'.env' file already exists"))
            filename = next(f"backup_{no}.env" for no in itertools.count() if not Path(f"backup_{no}.env").exists())

            if questionary.confirm(
                f"Do you want to overwrite the existing '.env' file? The existing will be renamed to {filename}",
                default=False,
            ).ask():
                shutil.move(".env", filename)
                Path(".env").write_text(new_env_file, encoding="utf-8")
        elif questionary.confirm("Do you want to save these to .env file for next time?", default=True).ask():
            Path(".env").write_text(new_env_file, encoding="utf-8")

    def verify(
        self,
        client: ToolkitClient,
        dry_run: bool,
        no_prompt: bool = False,
        demo_principal: str | None = None,
    ) -> VerifyAuthResult:
        """Authorization verification for the Toolkit.

        Args:
            client: The Toolkit client.
            dry_run: If the verification should be run in dry-run mode.
            no_prompt: If the verification should be run without any prompts.
            demo_principal: This is used for demo purposes. If passed, a different group name will be used
                to create the Toolkit group. This is group is intended to be deleted after the demo.

        Returns:
            VerifyAuthResult: The result of the verification.
        """

        is_interactive = not no_prompt
        is_demo = demo_principal is not None
        if client.config.project is None:
            raise AuthorizationError("CDF_PROJECT is not set.")
        cdf_project = client.config.project
        token_inspection = self.check_has_any_access(client)

        self.check_has_project_access(token_inspection, cdf_project)

        print(f"[italic]Focusing on current project {cdf_project} only from here on.[/]")

        self.check_has_group_access(client)

        self.check_identity_provider(client, cdf_project)

        try:
            user_groups = client.iam.groups.list()
        except CogniteAPIError as e:
            raise AuthorizationError(f"Unable to retrieve CDF groups.\n{e}")

        if not user_groups:
            raise AuthorizationError("The current user is not member of any groups in the CDF project.")

        loader_capabilities, loaders_by_capability_tuple = self._get_capabilities_by_loader(client)
        toolkit_group = self._create_toolkit_group(loader_capabilities, demo_principal)

        if not is_demo:
            print(
                Panel(
                    "The Cognite Toolkit expects the following:\n"
                    " - The principal used with the Toolkit [yellow]should[/yellow] be connected to "
                    "only ONE CDF Group.\n"
                    f" - This group [red]must[/red] be named {toolkit_group.name!r}.\n"
                    f" - The group {toolkit_group.name!r} [red]must[/red] have capabilities to "
                    f"all resources the Toolkit is managing\n"
                    " - All the capabilities [yellow]should[/yellow] be scoped to all resources.",
                    title="Toolkit Access Group",
                    expand=False,
                )
            )
            if is_interactive:
                Prompt.ask("Press enter key to continue...")

        all_groups = client.iam.groups.list(all=True)

        is_user_in_toolkit_group = any(group.name == toolkit_group.name for group in user_groups)
        is_toolkit_group_existing = any(group.name == toolkit_group.name for group in all_groups)

        print(f"Checking current client is member of the {toolkit_group.name!r} group...")
        has_added_capabilities = False
        cdf_toolkit_group: Group | None
        if is_user_in_toolkit_group:
            print(f"  [bold green]OK[/] - The current client is member of the {toolkit_group.name!r} group.")
            cdf_toolkit_group = next(group for group in user_groups if group.name == toolkit_group.name)
            missing_capabilities = self._check_missing_capabilities(
                client, cdf_toolkit_group, toolkit_group, loaders_by_capability_tuple, is_interactive
            )
            if (
                is_interactive
                and missing_capabilities
                and questionary.confirm("Do you want to update the group with the missing capabilities?").ask()
            ) or is_demo:
                has_added_capabilities = self._update_missing_capabilities(
                    client, cdf_toolkit_group, missing_capabilities, dry_run
                )
        elif is_toolkit_group_existing:  # and not is_user_in_toolkit_group
            self.warn(MediumSeverityWarning(f"The current client is not member of the {toolkit_group.name!r} group."))
            print(f"Checking if the group {toolkit_group.name!r} has the required capabilities...")
            # Update the group with the missing capabilities
            cdf_toolkit_group = next(group for group in all_groups if group.name == toolkit_group.name)
            missing_capabilities = self._check_missing_capabilities(
                client, cdf_toolkit_group, toolkit_group, loaders_by_capability_tuple, is_interactive
            )
            if (
                is_interactive
                and missing_capabilities
                and questionary.confirm("Do you want to update the group with the missing capabilities?").ask()
            ):
                self._update_missing_capabilities(client, cdf_toolkit_group, missing_capabilities, dry_run)
        elif is_demo:
            # We create the group for the demo user
            cdf_toolkit_group = self._create_toolkit_group_in_cdf(client, toolkit_group)
        else:
            print(f"Group {toolkit_group.name!r} does not exist in the CDF project.")
            cdf_toolkit_group = self._create_toolkit_group_in_cdf_interactive(
                client, toolkit_group, all_groups, is_interactive, dry_run
            )
        if cdf_toolkit_group is None:
            return VerifyAuthResult()

        if not is_demo and not is_user_in_toolkit_group:
            print(
                Panel(
                    "To use the Toolkit, for example, 'cdf deploy', [red]you need[/red] to make sure to use a service principal "
                    f"that is a member of the group with object id {cdf_toolkit_group.source_id!r}.",
                    title="Service Principal group membership",
                    expand=False,
                )
            )
            return VerifyAuthResult(function_status=None, toolkit_group_id=cdf_toolkit_group.id)

        if not is_demo:
            self.check_count_group_memberships(user_groups)

            self.check_source_id_usage(all_groups, cdf_toolkit_group)

            if extra := self.check_duplicated_names(all_groups, cdf_toolkit_group):
                if (
                    is_interactive
                    and questionary.confirm("Do you want to delete the extra groups?", default=True).ask()
                ):
                    try:
                        client.iam.groups.delete(extra.as_ids())
                    except CogniteAPIError as e:
                        raise ResourceDeleteError(f"Unable to delete the extra groups.\n{e}")
                    print(f"  [bold green]OK[/] - Deleted {len(extra)} duplicated groups.")

        function_status = self.check_function_service_status(client, dry_run, has_added_capabilities)
        return VerifyAuthResult(cdf_toolkit_group.id, function_status)

    def _create_toolkit_group_in_cdf_interactive(
        self,
        client: ToolkitClient,
        toolkit_group: GroupWrite,
        all_groups: GroupList,
        is_interactive: bool,
        dry_run: bool,
    ) -> Group | None:
        if not is_interactive:
            raise AuthorizationError(
                f"Group {toolkit_group.name!r} does not exist in the CDF project. "
                "Please create the group and try again."
                f"\n{HINT_LEAD_TEXT}Run this command without --no-prompt to get assistance to create the group."
            )
        if not questionary.confirm(
            "Do you want to create it?",
            default=True,
        ).ask():
            return None

        if dry_run:
            print(
                f"Would have created group {toolkit_group.name!r} with {len(toolkit_group.capabilities or [])} capabilities."
            )
            return None

        while True:
            source_id = questionary.text(
                "What is the source id for the new group (typically a group id in the identity provider)?"
            ).ask()
            if source_id:
                break
            print("Source id cannot be empty.")

        toolkit_group.source_id = source_id
        if already_used := [group.name for group in all_groups if group.source_id == source_id]:
            self.warn(
                HighSeverityWarning(
                    f"The source id {source_id!r} is already used by the groups: {humanize_collection(already_used)!r}."
                )
            )
            if not questionary.confirm("This is NOT recommended. Do you want to continue?", default=False).ask():
                return None

        return self._create_toolkit_group_in_cdf(client, toolkit_group)

    @staticmethod
    def _create_toolkit_group_in_cdf(
        client: ToolkitClient,
        toolkit_group: GroupWrite,
    ) -> Group:
        created = client.iam.groups.create(toolkit_group)
        print(
            f"  [bold green]OK[/] - Created new group {created.name}. It now has {len(created.capabilities or [])} capabilities."
        )
        return created

    def _check_missing_capabilities(
        self,
        client: ToolkitClient,
        existing_group: Group,
        toolkit_group: GroupWrite,
        loaders_by_capability_id: dict[tuple, list[str]],
        is_interactive: bool,
    ) -> list[Capability]:
        print(f"\nChecking if the {existing_group.name} has the all required capabilities...")
        with warnings.catch_warnings():
            # If the user has unknown capabilities, we don't want the user to see the warning:
            # "UserWarning: Unknown capability '<unknown warning>' will be ignored in comparison"
            # This is irrelevant for the user as we are only checking the capabilities below
            # (triggered by the verify_authorization calls)
            warnings.simplefilter("ignore")
            missing_capabilities = client.iam.compare_capabilities(
                existing_group.capabilities or [],
                toolkit_group.capabilities or [],
                project=client.config.project,
            )
        if not missing_capabilities:
            print(f"  [bold green]OK[/] - The {existing_group.name} has all the required capabilities.")
            return []

        missing_capabilities = self.merge_capabilities(missing_capabilities)
        for s in sorted(map(str, missing_capabilities)):
            self.warn(MissingCapabilityWarning(s))

        resource_names: set[str] = set()
        for cap in missing_capabilities:
            for cap_tuple in cap.as_tuples():
                resource_names.update(loaders_by_capability_id[cap_tuple])
        if resource_names:
            print("[bold yellow]INFO:[/] The missing capabilities are required for the following resources:")
            for resource_name in resource_names:
                print(f"    - {resource_name}")

        if not is_interactive:
            raise AuthorizationError(
                "The service principal/application does not have the required capabilities for the Toolkit to support all resources"
            )
        return missing_capabilities

    def _update_missing_capabilities(
        self,
        client: ToolkitClient,
        existing_group: Group,
        missing_capabilities: list[Capability],
        dry_run: bool,
    ) -> bool:
        """Updates the missing capabilities. This assumes interactive mode."""
        updated_toolkit_group = GroupWrite.load(existing_group.dump())
        if updated_toolkit_group.capabilities is None:
            updated_toolkit_group.capabilities = missing_capabilities
        else:
            updated_toolkit_group.capabilities.extend(missing_capabilities)

        with warnings.catch_warnings():
            # If the user has unknown capabilities, we don't want the user to see the warning:
            # "UserWarning: Unknown capability '<unknown warning>' will be ignored in comparison"
            # This is irrelevant for the user as we are only checking the capabilities below
            # (triggered by the verify_authorization calls)
            warnings.simplefilter("ignore")
            adding = client.iam.compare_capabilities(
                existing_group.capabilities or [],
                updated_toolkit_group.capabilities or [],
                project=client.config.project,
            )
        adding = self.merge_capabilities(adding)
        capability_str = "capabilities" if len(adding) > 1 else "capability"
        if dry_run:
            print(f"Would have updated group {updated_toolkit_group.name} with {len(adding)} new {capability_str}.")
            return False

        try:
            created = client.iam.groups.create(updated_toolkit_group)
        except CogniteAPIError as e:
            raise ResourceCreationError(f"Unable to create group {updated_toolkit_group.name}.\n{e}")
        try:
            client.iam.groups.delete(existing_group.id)
        except CogniteAPIError as e:
            raise ResourceDeleteError(
                f"Failed to cleanup old version of the {existing_group.name}.\n{e}\n"
                f"It is recommended that you manually delete the Group with ID {existing_group.id},"
                f"such that you don't have a duplicated group in your CDF project."
            )
        print(f"  [bold green]OK[/] - Updated the group {created.name} with {len(adding)} new {capability_str}.")
        return True

    @staticmethod
    def _create_toolkit_group(loader_capabilities: list[Capability], demo_user: str | None) -> GroupWrite:
        toolkit_group = GroupWrite(
            name=TOOLKIT_SERVICE_PRINCIPAL_GROUP_NAME if demo_user is None else TOOLKIT_DEMO_GROUP_NAME,
            capabilities=[
                *loader_capabilities,
                # Add project ACL to be able to list and read projects, as the Toolkit needs to know the project id.
                ProjectsAcl(
                    [ProjectsAcl.Action.Read, ProjectsAcl.Action.List, ProjectsAcl.Action.Update],
                    ProjectsAcl.Scope.All(),
                ),
                # Added Session ACL as this is required by CogIDP service principal
                SessionsAcl(
                    [SessionsAcl.Action.Create, SessionsAcl.Action.List, SessionsAcl.Action.Delete],
                    SessionsAcl.Scope.All(),
                ),
            ],
        )
        if demo_user:
            toolkit_group.members = [demo_user]
        return toolkit_group

    @staticmethod
    def _get_capabilities_by_loader(
        client: ToolkitClient,
    ) -> tuple[list[Capability], dict[tuple, list[str]]]:
        loaders_by_capability_tuple: dict[tuple, list[str]] = defaultdict(list)
        capability_by_id: dict[frozenset[tuple], Capability] = {}
        project_type = client.project.status().this_project.data_modeling_status
        for crud_cls in cruds.RESOURCE_CRUD_LIST:
            crud = crud_cls.create_loader(client)
            if crud.prerequisite_warning() is not None:
                continue
            capability = crud_cls.get_required_capability(None, read_only=False)
            capabilities = capability if isinstance(capability, list) else [capability]
            for cap in capabilities:
                if project_type == "DATA_MODELING_ONLY" and isinstance(cap, AssetsAcl | RelationshipsAcl):
                    continue
                id_ = frozenset(cap.as_tuples())
                if id_ not in capability_by_id:
                    capability_by_id[id_] = cap
                for cap_tuple in cap.as_tuples():
                    loaders_by_capability_tuple[cap_tuple].append(crud.display_name)
        return list(capability_by_id.values()), loaders_by_capability_tuple

    def check_has_any_access(self, client: ToolkitClient) -> TokenInspection:
        print("Checking basic project configuration...")
        try:
            # Using the token/inspect endpoint to check if the client has access to the project.
            # The response also includes access rights, which can be used to check if the client has the
            # correct access for what you want to do.
            token_inspection = client.iam.token.inspect()
            if token_inspection is None or len(token_inspection.capabilities) == 0:
                raise AuthorizationError(
                    "Valid authentication token, but it does not give any access rights."
                    " Check credentials (IDP_CLIENT_ID/IDP_CLIENT_SECRET or CDF_TOKEN)."
                )
            print("  [bold green]OK[/]")
        except CogniteAPIError as e:
            raise AuthorizationError(
                "Not a valid authentication token. Check credentials (IDP_CLIENT_ID/IDP_CLIENT_SECRET or CDF_TOKEN)."
                "This could also be due to the service principal/application not having access to any Groups."
                f"\n{e}"
            )
        return token_inspection

    def check_has_project_access(self, token_inspection: TokenInspection, cdf_project: str) -> None:
        print("Checking projects that the service principal/application has access to...")
        if len(token_inspection.projects) == 0:
            raise AuthorizationError(
                "The service principal/application configured for this client does not have access to any projects."
            )
        print("\n".join(f"  - {p.url_name}" for p in token_inspection.projects))
        if cdf_project not in {p.url_name for p in token_inspection.projects}:
            raise AuthorizationError(
                f"The service principal/application configured for this client does not have access to the CDF_PROJECT={cdf_project!r}."
            )

    def check_has_group_access(self, client: ToolkitClient) -> None:
        # Todo rewrite to use the token inspection instead.
        print(
            "Checking basic project and group manipulation access rights "
            "(projectsAcl: LIST, READ and groupsAcl: LIST, READ, CREATE, UPDATE, DELETE)..."
        )
        missing_capabilities = client.verify.authorization(
            [
                ProjectsAcl([ProjectsAcl.Action.List, ProjectsAcl.Action.Read], ProjectsAcl.Scope.All()),
                GroupsAcl(
                    [
                        GroupsAcl.Action.Read,
                        GroupsAcl.Action.List,
                        GroupsAcl.Action.Create,
                        GroupsAcl.Action.Update,
                        GroupsAcl.Action.Delete,
                    ],
                    GroupsAcl.Scope.All(),
                ),
            ]
        )
        if not missing_capabilities:
            print("  [bold green]OK[/]")
            return
        self.warn(
            HighSeverityWarning(
                "The service principal/application configured for this client "
                "does not have the basic group write access rights."
            )
        )
        print("Checking basic group read access rights (projectsAcl: LIST, READ and groupsAcl: LIST, READ)...")
        missing = client.verify.authorization(
            [
                ProjectsAcl([ProjectsAcl.Action.List, ProjectsAcl.Action.Read], ProjectsAcl.Scope.All()),
                GroupsAcl([GroupsAcl.Action.Read, GroupsAcl.Action.List], GroupsAcl.Scope.All()),
            ]
        )
        if not missing:
            print("  [bold green]OK[/] - can continue with checks.")
            return
        raise AuthorizationError(
            "Unable to continue, the service principal/application configured for this client does not"
            " have the basic read group access rights."
        )

    def check_identity_provider(self, client: ToolkitClient, cdf_project: str) -> None:
        print("Checking identity provider settings...")
        project_info = client.get(f"/api/v1/projects/{cdf_project}").json()
        oidc = project_info.get("oidcConfiguration", {})
        if "https://login.windows.net" in oidc.get("tokenUrl"):
            tenant_id = oidc.get("tokenUrl").split("/")[-3]
            print(f"  [bold green]OK[/]: Microsoft Entra ID (aka ActiveDirectory) with tenant id ({tenant_id}).")
        elif "auth0.com" in oidc.get("tokenUrl"):
            tenant_id = oidc.get("tokenUrl").split("/")[2].split(".")[0]
            print(f"  [bold green]OK[/] - Auth0 with tenant id ({tenant_id}).")
        else:
            self.warn(MediumSeverityWarning(f"Unknown identity provider {oidc.get('tokenUrl')}"))
        access_claims = [c.get("claimName") for c in oidc.get("accessClaims", {})]
        print(
            f"  Matching on CDF group sourceIds will be done on any of these claims from the identity provider: {access_claims}"
        )

    def check_count_group_memberships(self, user_group: GroupList) -> None:
        print("Checking CDF group memberships for the current client configured...")

        table = Table(title="CDF Group ids, Names, and Source Ids")
        table.add_column("Id", justify="left")
        table.add_column("Name", justify="left")
        table.add_column("Source Id", justify="left")
        for group in user_group:
            name = group.name
            if group.name == TOOLKIT_SERVICE_PRINCIPAL_GROUP_NAME:
                name = f"[bold]{group.name}[/]"
            table.add_row(str(group.id), name, group.source_id)
        print(table)

        if len(user_group) > 1:
            self.warn(
                LowSeverityWarning(
                    "This service principal/application gets its access rights from more than one CDF group."
                    "\nThis is not recommended. The group matching the group config file is marked in "
                    "bold above if it is present."
                )
            )
        else:
            print("  [bold green]OK[/] - Only one group is used for this service principal/application.")

    def check_source_id_usage(self, all_groups: GroupList, cdf_toolkit_group: Group) -> None:
        reuse_source_id = [
            group.name
            for group in all_groups
            if group.source_id == cdf_toolkit_group.source_id and group.id != cdf_toolkit_group.id
        ]
        if reuse_source_id:
            group_names_str = humanize_collection(reuse_source_id)
            self.warn(
                MediumSeverityWarning(
                    f"The following groups have the same source id, {cdf_toolkit_group.source_id},\n"
                    f"as the {cdf_toolkit_group.name!r} group: \n    {group_names_str!r}.\n"
                    f"It is recommended that only the {cdf_toolkit_group.name!r} group has this source id."
                )
            )

    def check_duplicated_names(self, all_groups: GroupList, cdf_toolkit_group: Group) -> GroupList:
        extra = GroupList(
            [group for group in all_groups if group.name == cdf_toolkit_group.name and group.id != cdf_toolkit_group.id]
        )
        if extra:
            self.warn(
                MediumSeverityWarning(
                    f"There are multiple groups with the same name {cdf_toolkit_group.name} in the CDF project."
                    "           It is recommended that this admin (CI/CD) application/service principal "
                    "only is member of one group in the identity provider. Suggest you delete all but one"
                    "           of the groups with the same name."
                )
            )

        return extra

    @staticmethod
    def merge_capabilities(capability_list: list[Capability]) -> list[Capability]:
        """Merges capabilities that have the same ACL and Scope"""
        actions_by_scope_and_cls: dict[tuple[type[Capability], frozenset[tuple]], set[Capability.Action]] = defaultdict(
            set
        )
        for capability in capability_list:
            scope_key = frozenset(capability.scope.as_tuples())
            actions_by_scope_and_cls[(type(capability), scope_key)].update(capability.actions)

        scope_map: dict[tuple[type[Capability], frozenset[tuple]], Capability.Scope] = {
            (type(cap), frozenset(cap.scope.as_tuples())): cap.scope for cap in capability_list
        }

        return [
            cap_cls(actions=list(actions), scope=scope_map[(cap_cls, scope)], allow_unknown=False)
            for (cap_cls, scope), actions in actions_by_scope_and_cls.items()
        ]

    def check_function_service_status(
        self, client: ToolkitClient, dry_run: bool, has_added_capabilities: bool
    ) -> str | None:
        print("Checking function service status...")
        has_function_read_access = self.has_function_rights(client, [FunctionsAcl.Action.Read], has_added_capabilities)
        if not has_function_read_access:
            self.warn(HighSeverityWarning("Cannot check function service status, missing function read access."))
            return None
        try:
            function_status = client.functions.status()
        except CogniteAPIError as e:
            self.warn(HighSeverityWarning(f"Unable to check function service status.\n{e}"))
            return None

        if function_status.status == "requested":
            print("  [bold yellow]INFO:[/] Function service activation is in progress (may take up to 2 hours)...")
        elif dry_run and function_status.status != "activated":
            print(
                "  [bold yellow]INFO:[/] Function service has not been activated, "
                "would have activated (will take up to 2 hours)..."
            )
        elif not dry_run and function_status.status != "activated":
            has_function_write_access = self.has_function_rights(
                client, [FunctionsAcl.Action.Write], has_added_capabilities
            )
            if not has_function_write_access:
                self.warn(HighSeverityWarning("Cannot activate function service, missing function write access."))
                return function_status.status

            if client.config.is_private_link:
                print(
                    "  [bold yellow]INFO:[/] Function service has not been activated. "
                    "Function activation must be done manually."
                )
                return function_status.status
            try:
                client.functions.activate()
            except CogniteAPIError as e:
                self.warn(HighSeverityWarning(f"Unable to activate function service.\n{e}"))
                return function_status.status
            print(
                "  [bold green]OK[/] - Function service has been activated. This may take up to 2 hours to take effect."
            )
        else:
            print("  [bold green]OK[/] - Function service has been activated.")

        return function_status.status

    def has_function_rights(
        self, client: ToolkitClient, actions: list[FunctionsAcl.Action], has_added_capabilities: bool
    ) -> bool:
        t0 = time.perf_counter()
        while not (
            has_function_access := not client.iam.verify_capabilities(
                FunctionsAcl(actions, FunctionsAcl.Scope.All()),
            )
        ):
            if has_added_capabilities and (time.perf_counter() - t0 < 5.0):
                # Wait for the IAM service to update the capabilities
                sleep(1.0)
            else:
                break
        return has_function_access



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import (
    ClientCredentials,
    OidcCredentials,
    Transformation,
    TransformationList,
    TransformationNotification,
    TransformationNotificationList,
    TransformationSchedule,
    TransformationScheduleList,
    TransformationScheduleWrite,
    TransformationScheduleWriteList,
    TransformationWrite,
    TransformationWriteList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    TransformationsAcl,
)
from cognite.client.data_classes.data_modeling.ids import (
    DataModelId,
    ViewId,
)
from cognite.client.data_classes.transformations import NonceCredentials
from cognite.client.data_classes.transformations.notifications import (
    TransformationNotificationWrite,
    TransformationNotificationWriteList,
)
from cognite.client.exceptions import CogniteAPIError, CogniteAuthError, CogniteDuplicatedError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ToolkitFileNotFoundError,
    ToolkitInvalidParameterNameError,
    ToolkitNotSupported,
    ToolkitRequiredValueError,
    ToolkitYAMLFormatError,
)
from cognite_toolkit._cdf_tk.resource_classes import (
    TransformationNotificationYAML,
    TransformationScheduleYAML,
    TransformationYAML,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    calculate_secure_hash,
    humanize_collection,
    in_dict,
    load_yaml_inject_variables,
    quote_int_value_by_key_in_yaml,
    safe_read,
)
from cognite_toolkit._cdf_tk.utils.cdf import read_auth, try_find_error
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD
from .datamodel import DataModelCRUD, SpaceCRUD, ViewCRUD
from .group_scoped import GroupResourceScopedCRUD
from .raw import RawDatabaseCRUD, RawTableCRUD


@final
class TransformationCRUD(
    ResourceCRUD[str, TransformationWrite, Transformation, TransformationWriteList, TransformationList]
):
    folder_name = "transformations"
    filename_pattern = (
        # Matches all yaml files except file names whose stem contain *.schedule. or .Notification
        r"^(?!.*schedule.*|.*\.notification$).*$"
    )
    resource_cls = Transformation
    resource_write_cls = TransformationWrite
    list_cls = TransformationList
    list_write_cls = TransformationWriteList
    kind = "Transformation"
    yaml_cls = TransformationYAML
    dependencies = frozenset(
        {
            DataSetsCRUD,
            RawDatabaseCRUD,
            GroupAllScopedCRUD,
            SpaceCRUD,
            ViewCRUD,
            DataModelCRUD,
            RawTableCRUD,
            RawDatabaseCRUD,
            GroupResourceScopedCRUD,
        }
    )
    _doc_url = "Transformations/operation/createTransformations"
    _hash_key = "-- cdf-auth"

    # The API supports creating/updating 1000 transformations in a single request,
    # however, when the transformation has credentials, the session API times out on
    # larger number of transformations. Thus, we use a conservative batch size.
    _BATCH_SIZE = 20  # The maximum number of transformations to create in a single batch

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._authentication_by_id_operation: dict[
            tuple[str, Literal["read", "write"]], OidcCredentials | ClientCredentials
        ] = {}

    @property
    def display_name(self) -> str:
        return "transformations"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [TransformationsAcl.Action.Read]
            if read_only
            else [TransformationsAcl.Action.Read, TransformationsAcl.Action.Write]
        )
        scope: TransformationsAcl.Scope.All | TransformationsAcl.Scope.DataSet = TransformationsAcl.Scope.All()  # type: ignore[valid-type]
        if items is not None:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TransformationsAcl.Scope.DataSet(list(data_set_ids))

        return TransformationsAcl(actions, scope)

    @classmethod
    def get_id(cls, item: Transformation | TransformationWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("Transformation must have external_id set.")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: Transformation | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if item.id is None:
            raise ToolkitRequiredValueError("Transformation must have id set.")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        if destination := item.get("destination", {}):
            if not isinstance(destination, dict):
                return
            if destination.get("type") == "raw" and in_dict(("database", "table"), destination):
                yield RawDatabaseCRUD, RawDatabase(destination["database"])
                yield RawTableCRUD, RawTable(destination["database"], destination["table"])
            elif destination.get("type") in ("nodes", "edges") and (view := destination.get("view", {})):
                if space := destination.get("instanceSpace"):
                    yield SpaceCRUD, space
                if in_dict(("space", "externalId", "version"), view):
                    view["version"] = str(view["version"])
                    yield ViewCRUD, ViewId.load(view)
            elif destination.get("type") == "instances":
                if space := destination.get("instanceSpace"):
                    yield SpaceCRUD, space
                if data_model := destination.get("dataModel"):
                    if in_dict(("space", "externalId", "version"), data_model):
                        data_model["version"] = str(data_model["version"])
                        yield DataModelCRUD, DataModelId.load(data_model)

    def safe_read(self, filepath: Path | str) -> str:
        # If the destination is a DataModel or a View we need to ensure that the version is a string
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = load_yaml_inject_variables(
            self.safe_read(filepath),
            environment_variables or {},
            original_filepath=filepath,
        )

        raw_list = resources if isinstance(resources, list) else [resources]
        for item in raw_list:
            query_file: Path | None = None
            if "queryFile" in item:
                if filepath is None:
                    raise ValueError("filepath must be set if queryFile is set")
                query_file = filepath.parent / Path(item.pop("queryFile"))

            external_id = self.get_id(item)
            if query_file is None and "query" not in item:
                if filepath is None:
                    raise ValueError("filepath must be set if query is not set")
                warning = HighSeverityWarning(
                    f"query property or is missing in {filepath.as_posix()!r}. It can be inline or a separate file named {filepath.stem}.sql or {external_id}.sql",
                )
                warning.print_warning(console=self.console)
            elif query_file and not query_file.exists():
                # We checked above that filepath is not None
                raise ToolkitFileNotFoundError(f"Query file {query_file.as_posix()} not found", filepath)
            elif query_file and "query" in item:
                raise ToolkitYAMLFormatError(
                    f"query property is ambiguously defined in both the yaml file and a separate file named {query_file}\n"
                    f"Please remove one of the definitions, either the query property in {filepath} or the file {query_file}",
                    filepath,
                )
            elif query_file:
                item["query"] = safe_read(query_file, encoding=BUILD_FOLDER_ENCODING)

            auth_dict: dict[str, Any] = {}
            if "authentication" in item:
                auth_dict["authentication"] = item["authentication"]
            if auth_dict:
                auth_hash = calculate_secure_hash(auth_dict, shorten=True)
                if "query" in item:
                    hash_str = f"{self._hash_key}: {auth_hash}"
                    if not item["query"].startswith(self._hash_key):
                        item["query"] = f"{hash_str}\n{item['query']}"

            if "sourceOidcCredentials" in item:
                raise ToolkitNotSupported(
                    "The property 'sourceOidcCredentials' is not supported. Use 'authentication.read' instead."
                )

            if "destinationOidcCredentials" in item:
                raise ToolkitNotSupported(
                    "The property 'destinationOidcCredentials' is not supported. Use 'authentication.write' instead."
                )

            if "sourceNonce" in item:
                raise ToolkitNotSupported(
                    "The property 'sourceNonce' is not supported by Toolkit. Use 'authentication.read' instead,"
                    "then Toolkit will dynamically set the nonce."
                )

            if "destinationNonce" in item:
                raise ToolkitNotSupported(
                    "The property 'destinationNonce' is not supported by Toolkit. Use 'authentication.write' instead,"
                    "then Toolkit will dynamically set the nonce."
                )
            auth = item.pop("authentication", None)
            if isinstance(auth, dict) and "read" in auth:
                self._authentication_by_id_operation[(external_id, "read")] = read_auth(
                    auth["read"],
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            elif isinstance(auth, dict) and "write" not in auth:
                self._authentication_by_id_operation[(external_id, "read")] = read_auth(
                    auth,
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            if isinstance(auth, dict) and "write" in auth:
                self._authentication_by_id_operation[(external_id, "write")] = read_auth(
                    auth["write"],
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            elif isinstance(auth, dict) and "read" not in auth:
                self._authentication_by_id_operation[(external_id, "write")] = read_auth(
                    auth,
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
        return raw_list

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> TransformationWrite:
        invalid_parameters: dict[str, str] = {}
        if "action" in resource and "conflictMode" not in resource:
            invalid_parameters["action"] = "conflictMode"
        if "shared" in resource and "isPublic" not in resource:
            invalid_parameters["shared"] = "isPublic"
        if invalid_parameters:
            raise ToolkitInvalidParameterNameError(
                "Parameters invalid. These are specific for the "
                "'transformation-cli' and not supported by cognite-toolkit",
                resource.get("externalId", "<Missing>"),
                invalid_parameters,
            )

        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if "conflictMode" not in resource:
            # Todo; Bug SDK missing default value
            resource["conflictMode"] = "upsert"
        return TransformationWrite._load(resource)

    def dump_resource(self, resource: Transformation, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if "isPublic" in dumped and "isPublic" not in local:
            # Default set from server side.
            dumped.pop("isPublic")
        if "authentication" in local:
            # The hash added to the beginning of the query detects the change in the authentication
            dumped["authentication"] = local["authentication"]
        return dumped

    def split_resource(
        self, base_filepath: Path, resource: dict[str, Any]
    ) -> Iterable[tuple[Path, dict[str, Any] | str]]:
        if query := resource.pop("query", None):
            yield base_filepath.with_suffix(".sql"), cast(str, query)

        yield base_filepath, resource

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[-1] == "scopes":
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[TransformationWrite]) -> TransformationList:
        return self._execute_in_batches(items, self.client.transformations.create)

    def retrieve(self, ids: SequenceNotStr[str | int]) -> TransformationList:
        internal_ids, external_ids = self._split_ids(ids)
        return self.client.transformations.retrieve_multiple(
            ids=internal_ids, external_ids=external_ids, ignore_unknown_ids=True
        )

    def update(self, items: Sequence[TransformationWrite]) -> TransformationList:
        def update(transformations: Sequence[TransformationWrite]) -> TransformationList:
            return self.client.transformations.update(transformations, mode="replace")

        return self._execute_in_batches(items, update)

    @staticmethod
    def _create_auth_creation_error(items: Sequence[TransformationWrite]) -> ResourceCreationError | None:
        hints_by_id: dict[str, list[str]] = defaultdict(list)
        for item in items:
            if not item.external_id:
                continue
            hint_source = try_find_error(item.source_oidc_credentials)
            if hint_source:
                hints_by_id[item.external_id].append(hint_source)
            if hint_dest := try_find_error(item.destination_oidc_credentials):
                if hint_dest != hint_source:
                    hints_by_id[item.external_id].append(hint_dest)
        if hints_by_id:
            body = "\n".join(f"  {id_} - {humanize_collection(hints)}" for id_, hints in hints_by_id.items())
            return ResourceCreationError(
                f"Failed to create Transformation(s) du to likely invalid credentials:\n{body}",
            )
        return None

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        existing = self.retrieve(ids).as_ids()
        if existing:
            self.client.transformations.delete(id=existing, ignore_unknown_ids=True)
        return len(existing)

    def _execute_in_batches(
        self,
        items: Sequence[TransformationWrite],
        api_call: Callable[[Sequence[TransformationWrite]], TransformationList],
    ) -> TransformationList:
        results = TransformationList([])
        for chunk in chunker(items, self._BATCH_SIZE):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._update_nonce(chunk)
            try:
                chunk_results = api_call(chunk)
            except CogniteAuthError as e:
                if error := self._create_auth_creation_error(chunk):
                    raise error from e
                raise e
            results.extend(chunk_results)
        return results

    def _update_nonce(self, items: Sequence[TransformationWrite]) -> None:
        for item in items:
            if not item.external_id:
                raise ToolkitRequiredValueError("Transformation must have external_id set.")
            if read_credentials := self._authentication_by_id_operation.get((item.external_id, "read")):
                item.source_nonce = self._create_nonce(read_credentials)
            if write_credentials := self._authentication_by_id_operation.get((item.external_id, "write")):
                item.destination_nonce = self._create_nonce(write_credentials)

    def _create_nonce(self, credentials: OidcCredentials | ClientCredentials) -> NonceCredentials:
        if isinstance(credentials, ClientCredentials):
            session = self.client.iam.sessions.create(credentials)
            nonce = NonceCredentials(session.id, session.nonce, self.client.config.project)
        elif isinstance(credentials, OidcCredentials):
            config = deepcopy(self.client.config)
            config.project = credentials.cdf_project_name
            config.credentials = credentials.as_credential_provider()
            other_client = ToolkitClient(config)
            session = other_client.iam.sessions.create(credentials.as_client_credentials())
            nonce = NonceCredentials(session.id, session.nonce, credentials.cdf_project_name)
        else:
            raise ValueError(f"Error in TransformationLoader: {type(credentials)} is not a valid credentials type")
        return nonce

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Transformation]:
        return iter(
            self.client.transformations(data_set_external_ids=[data_set_external_id] if data_set_external_id else None)
        )

    def sensitive_strings(self, item: TransformationWrite) -> Iterable[str]:
        if item.source_oidc_credentials:
            yield item.source_oidc_credentials.client_secret
        if item.destination_oidc_credentials:
            yield item.destination_oidc_credentials.client_secret

        external_id = self.get_id(item)
        if read_credentials := self._authentication_by_id_operation.get((external_id, "read")):
            yield read_credentials.client_secret
        if write_credentials := self._authentication_by_id_operation.get((external_id, "write")):
            yield write_credentials.client_secret


@final
class TransformationScheduleCRUD(
    ResourceCRUD[
        str,
        TransformationScheduleWrite,
        TransformationSchedule,
        TransformationScheduleWriteList,
        TransformationScheduleList,
    ]
):
    folder_name = "transformations"
    # Matches all yaml files whose stem contains *schedule or *TransformationSchedule.
    filename_pattern = r"^.*schedule$"
    resource_cls = TransformationSchedule
    resource_write_cls = TransformationScheduleWrite
    list_cls = TransformationScheduleList
    list_write_cls = TransformationScheduleWriteList
    kind = "Schedule"
    yaml_cls = TransformationScheduleYAML
    dependencies = frozenset({TransformationCRUD})
    _doc_url = "Transformation-Schedules/operation/createTransformationSchedules"
    parent_resource = frozenset({TransformationCRUD})

    @property
    def display_name(self) -> str:
        return "transformation schedules"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationScheduleWrite] | None, read_only: bool
    ) -> list[Capability]:
        # Access for transformations schedules is checked by the transformation that is deployed
        # first, so we don't need to check for any capabilities here.
        return []

    @classmethod
    def get_id(cls, item: TransformationSchedule | TransformationScheduleWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("TransformationSchedule must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "externalId" in item:
            yield TransformationCRUD, item["externalId"]

    def create(self, items: Sequence[TransformationScheduleWrite]) -> TransformationScheduleList:
        try:
            return self.client.transformations.schedules.create(list(items))
        except CogniteDuplicatedError as e:
            existing = {external_id for dup in e.duplicated if (external_id := dup.get("externalId", None))}
            print(
                f"  [bold yellow]WARNING:[/] {len(e.duplicated)} transformation schedules already exist(s): {existing}"
            )
            new_items = [item for item in items if item.external_id not in existing]
            return self.client.transformations.schedules.create(new_items)

    def retrieve(self, ids: SequenceNotStr[str]) -> TransformationScheduleList:
        return self.client.transformations.schedules.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: TransformationScheduleWriteList) -> TransformationScheduleList:
        return self.client.transformations.schedules.update(items, mode="replace")

    def delete(self, ids: str | SequenceNotStr[str] | None) -> int:
        try:
            self.client.transformations.schedules.delete(
                external_id=cast(SequenceNotStr[str], ids), ignore_unknown_ids=False
            )
            return len(cast(SequenceNotStr[str], ids))
        except CogniteNotFoundError as e:
            return len(cast(SequenceNotStr[str], ids)) - len(e.not_found)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[TransformationSchedule]:
        if parent_ids is None:
            yield from iter(self.client.transformations.schedules)
        else:
            for transformation_id in parent_ids:
                if isinstance(transformation_id, str):
                    res = self.client.transformations.schedules.retrieve(external_id=transformation_id)
                    if res:
                        yield res
                elif isinstance(transformation_id, int):
                    res = self.client.transformations.schedules.retrieve(id=transformation_id)
                    if res:
                        yield res


@final
class TransformationNotificationCRUD(
    ResourceCRUD[
        str,
        TransformationNotificationWrite,
        TransformationNotification,
        TransformationNotificationWriteList,
        TransformationNotificationList,
    ]
):
    folder_name = "transformations"
    # Matches all yaml files whose stem ends with *Notification.
    filename_pattern = r"^.*Notification$"
    resource_cls = TransformationNotification
    resource_write_cls = TransformationNotificationWrite
    list_cls = TransformationNotificationList
    list_write_cls = TransformationNotificationWriteList
    kind = "Notification"
    dependencies = frozenset({TransformationCRUD})
    _doc_url = "Transformation-Notifications/operation/createTransformationNotifications"
    _split_character = "@@@"
    parent_resource = frozenset({TransformationCRUD})
    yaml_cls = TransformationNotificationYAML

    @property
    def display_name(self) -> str:
        return "transformation notifications"

    @classmethod
    def get_id(cls, item: TransformationNotification | TransformationNotificationWrite | dict) -> str:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"transformationExternalId", "destination"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return f"{item['transformationExternalId']}{cls._split_character}{item['destination']}"

        return f"{item.transformation_external_id}{cls._split_character}{item.destination}"

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        transformation_id, destination = id.split(cls._split_character, maxsplit=1)
        return {
            "transformationExternalId": transformation_id,
            "destination": destination,
        }

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationNotificationWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # Access for transformation notification is checked by the transformation that is deployed
        # first, so we don't need to check for any capabilities here.
        return []

    def dump_resource(
        self, resource: TransformationNotification, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if local and "transformationExternalId" in local:
            dumped.pop("transformationId")
            dumped["transformationExternalId"] = local["transformationExternalId"]
        return dumped

    def create(self, items: TransformationNotificationWriteList) -> TransformationNotificationList:
        return self.client.transformations.notifications.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> TransformationNotificationList:
        retrieved = TransformationNotificationList([])
        for id_ in ids:
            try:
                transformation_external_id, destination = id_.rsplit(self._split_character, maxsplit=1)
            except ValueError:
                # This should never happen, and is a bug in the toolkit if it occurs. Creating a nice error message
                # here so that if it does happen, it will be easier to debug.
                raise ValueError(
                    f"Invalid externalId: {id_}. Must be in the format 'transformationExternalId{self._split_character}destination'"
                )
            try:
                result = self.client.transformations.notifications.list(
                    transformation_external_id=transformation_external_id, destination=destination, limit=-1
                )
                # list() does not return the transformation_external_id on items
                for notification in result:
                    notification.transformation_external_id = transformation_external_id

            except CogniteAPIError:
                # The notification endpoint gives a 500 if the notification does not exist.
                # The issue has been reported to the service team.
                continue

            retrieved.extend(result)

        return retrieved

    def update(self, items: TransformationNotificationWriteList) -> TransformationNotificationList:
        # Note that since a notification is identified by the combination of transformationExternalId and destination,
        # which is the entire object, an update should never happen. However, implementing just in case.
        item_by_id = {self.get_id(item): item for item in items}
        existing = self.retrieve(list(item_by_id.keys()))
        exiting_by_id = {self.get_id(item): item for item in existing}
        create: list[TransformationNotificationWrite] = []
        unchanged: list[str] = []
        delete: list[int] = []
        for id_, local_item in item_by_id.items():
            existing_item = exiting_by_id.get(id_)
            local_dict = local_item.dump()
            if existing_item and local_item == self.dump_resource(existing_item, local_dict):
                unchanged.append(self.get_id(existing_item))
            else:
                create.append(local_item)
            if existing_item:
                delete.append(existing_item.id)
        if delete:
            self.client.transformations.notifications.delete(delete)
        updated_by_id: dict[str, TransformationNotification] = {}
        if create:
            # Bug in SDK
            created = self.client.transformations.notifications.create(create)
            updated_by_id.update({self.get_id(item): item for item in created})
        if unchanged:
            updated_by_id.update({id_: exiting_by_id[id_] for id_ in unchanged})
        return TransformationNotificationList([updated_by_id[id_] for id_ in item_by_id.keys()])

    def delete(self, ids: SequenceNotStr[str]) -> int:
        # Note that it is theoretically possible that more items will be deleted than
        # input ids. This is because TransformationNotifications are identified by an internal id,
        # while the toolkit uses the transformationExternalId + destination as the id. Thus, there could
        # be multiple notifications for the same transformationExternalId + destination.
        if existing := self.retrieve(ids):
            self.client.transformations.notifications.delete([item.id for item in existing])
        return len(existing)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[TransformationNotification]:
        if parent_ids is None:
            yield from iter(self.client.transformations.notifications)
        else:
            for transformation_id in parent_ids:
                if isinstance(transformation_id, str):
                    yield from self.client.transformations.notifications(transformation_external_id=transformation_id)
                elif isinstance(transformation_id, int):
                    yield from self.client.transformations.notifications(transformation_id=transformation_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "transformationExternalId" in item:
            yield TransformationCRUD, item["transformationExternalId"]



from typing import Literal

from pydantic import Field, model_validator

from .base import BaseModelResource, ToolkitResource


class SequenceColumnDTO(BaseModelResource):
    external_id: str = Field(
        description="The external ID of the column.",
        max_length=255,
    )
    name: str | None = Field(
        default=None,
        description="The name of the column.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="The description of the column.",
        max_length=1000,
    )
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Custom, application-specific metadata.",
        max_length=256,
    )
    value_type: Literal["STRING", "string", "DOUBLE", "double", "LONG", "long"] | None = Field(
        default=None,
        description="What type the datapoints in a column will have.",
    )


class SequenceYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    name: str | None = Field(
        default=None,
        description="The name of the sequence.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="The description of the sequence.",
        max_length=1000,
    )
    data_set_external_id: str | None = Field(
        default=None,
        description="The external ID of the data set that the sequence associated with.",
    )
    asset_external_id: str | None = Field(
        default=None,
        description="The external ID of the asset that the sequence associated with.",
    )
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Custom, application-specific metadata.",
        max_length=256,
    )
    columns: list[SequenceColumnDTO] = Field(
        description="List of column definitions.",
        min_length=1,
        max_length=400,
    )


class SequenceRowDTO(BaseModelResource):
    row_number: int = Field(
        description="The row number of the row.",
        ge=0,
    )
    values: list[str | int | float | None] = Field(
        description="List of values in the order defined in the columns field",
        min_length=1,
        max_length=400,
    )


class SequenceRowYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=256,
    )
    columns: list[str] = Field(
        description="List of column definitions.",
        min_length=1,
        max_length=200,
    )
    rows: list[SequenceRowDTO] = Field(
        description="List of row definitions.",
        min_length=1,
        max_length=10000,
    )

    @model_validator(mode="after")
    def validate_values_match_columns(self) -> "SequenceRowYAML":
        """Validate that the number of values in each row matches the number of columns."""
        total_columns = len(self.columns)
        for i, row in enumerate(self.rows):
            if len(row.values) != total_columns:
                raise ValueError(
                    f"Row number {row.row_number} has {len(row.values)} value(s). "
                    f"Each row must have exactly {total_columns} value(s) which is the same as the number of column(s)."
                )
        return self



import pytest
from cognite.client.data_classes.data_modeling import SpaceApply

from cognite_toolkit._cdf_tk.client import ToolkitClient


@pytest.fixture(scope="session")
def dev_space(dev_cluster_client: ToolkitClient) -> str:
    """Fixture to create a space for the tests."""
    space_name = "toolkit_test_space"
    dev_cluster_client.data_modeling.spaces.apply(SpaceApply(space=space_name))
    return space_name



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.extraction_pipeline import ExtractionPipelineYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_extraction_pipeline_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "myPipeline"},
        {"Missing required field: 'name'", "Missing required field: 'dataSetExternalId'"},
        id="Missing required fields",
    )
    # Missing externalId
    yield pytest.param(
        {"name": "Pipeline 2", "dataSetExternalId": "ds1"},
        {"Missing required field: 'externalId'"},
        id="Missing externalId",
    )
    # Empty name
    yield pytest.param(
        {"externalId": "pipeline3", "name": "", "dataSetExternalId": "ds2"},
        {"In field name string should have at least 1 character"},
        id="Empty name",
    )
    # Invalid dataSetExternalId type
    yield pytest.param(
        {"externalId": "pipeline4", "name": "Pipeline 4", "dataSetExternalId": 123},
        {
            "In field dataSetExternalId input should be a valid string. Got 123 of type "
            "int. Hint: Use double quotes to force string."
        },
        id="Invalid dataSetExternalId type",
    )
    # All required fields present but with extra unknown field
    yield pytest.param(
        {"externalId": "pipeline5", "name": "Pipeline 5", "dataSetExternalId": "ds5", "unknownField": "value"},
        {"Unused field: 'unknownField'"},
        id="Unknown field present",
    )


class TestExtractionPipelineYAML:
    @pytest.mark.parametrize(
        "data",
        [
            *find_resources("ExtractionPipeline"),
            {
                "externalId": "pipeline6",
                "name": "Pipeline 6",
                "dataSetExternalId": "ds6",
                "contacts": [{"name": "John Doe", "role": "Owner", "sendNotification": True}],
                "notificationConfig": {"allowedNotSeenRangeInMinutes": 10},
            },
        ],
    )
    def test_load_valid_extraction_pipeline(self, data: dict[str, object]) -> None:
        loaded = ExtractionPipelineYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_extraction_pipeline_test_cases()))
    def test_invalid_extraction_pipeline_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, ExtractionPipelineYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import pytest
from cognite.client.data_classes.data_modeling.ids import DataModelId

from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    AssetCentricFilter,
    AssetCentricSubFilter,
    LocationFilterScene,
    LocationFilterWrite,
)
from cognite_toolkit._cdf_tk.cruds._resource_cruds.location import LocationFilterCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitCycleError
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import LOAD_DATA
from tests.test_unit.approval_client.client import ApprovalToolkitClient


@pytest.fixture
def exhaustive_filter(env_vars_with_client: EnvironmentVariables) -> LocationFilterWrite:
    loader = LocationFilterCRUD.create_loader(env_vars_with_client.get_client())
    raw_list = loader.load_resource_file(
        LOAD_DATA / "locations" / "exhaustive.LocationFilter.yaml", env_vars_with_client.dump()
    )
    loaded = loader.load_resource(raw_list[0], is_dry_run=False)
    return loaded


class TestLocationFilterLoader:
    def test_load_minimum_location_filter(
        self,
        env_vars_with_client: EnvironmentVariables,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        loader = LocationFilterCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(
            LOAD_DATA / "locations" / "minimum.LocationFilter.yaml", env_vars_with_client.dump()
        )
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)
        assert isinstance(loaded, LocationFilterWrite)
        assert loaded.external_id == "springfield"

    def test_load_filter_write(self, exhaustive_filter: LocationFilterWrite) -> None:
        assert isinstance(exhaustive_filter, LocationFilterWrite)
        assert exhaustive_filter.external_id == "unique-external-id-123"

    def test_load_filter_write_data_models(self, exhaustive_filter: LocationFilterWrite) -> None:
        assert isinstance(exhaustive_filter.data_models[0], DataModelId)
        assert exhaustive_filter.data_models[0].external_id == "data-model-id-456"

    def test_load_filter_write_instance_spaces(self, exhaustive_filter: LocationFilterWrite) -> None:
        assert isinstance(exhaustive_filter.instance_spaces, list)
        assert exhaustive_filter.instance_spaces[0] == "instance-space-main"

    def test_load_filter_write_scene(self, exhaustive_filter: LocationFilterWrite) -> None:
        assert isinstance(exhaustive_filter.scene, LocationFilterScene)
        assert exhaustive_filter.scene.external_id == "scene-id-012"

    def test_load_filter_write_asset_centric(self, exhaustive_filter: LocationFilterWrite) -> None:
        assert isinstance(exhaustive_filter.asset_centric, AssetCentricFilter)
        assert isinstance(exhaustive_filter.asset_centric.assets, AssetCentricSubFilter)
        assert exhaustive_filter.asset_centric.asset_subtree_ids[0] == {"externalId": "general-subtree-id-890"}
        assert exhaustive_filter.asset_centric.assets.asset_subtree_ids[0] == {"externalId": "root-asset"}
        assert exhaustive_filter.asset_centric.events.asset_subtree_ids[0] == {"externalId": "event-subtree-id-678"}

    def test_topological_sort_success(self) -> None:
        # Create location filters with parent-child relationships
        # Structure: grandparent -> parent -> child
        location_filters = [
            LocationFilterWrite(
                external_id="child",
                name="Child Location",
                parent_id=-1,  # Deferred lookup
                _parent_external_id="parent",
            ),
            LocationFilterWrite(
                external_id="parent",
                name="Parent Location",
                parent_id=-1,  # Deferred lookup
                _parent_external_id="grandparent",
            ),
            LocationFilterWrite(
                external_id="grandparent",
                name="Grandparent Location",
                parent_id=None,  # No parent
            ),
        ]

        sorted_filters = LocationFilterCRUD.topological_sort(location_filters)

        # Should be sorted with grandparent first, then parent, then child
        assert len(sorted_filters) == 3
        assert sorted_filters[0].external_id == "grandparent"
        assert sorted_filters[1].external_id == "parent"
        assert sorted_filters[2].external_id == "child"

    def test_topological_sort_raises_on_cycle(self) -> None:
        # Create location filters with circular dependencies
        dependencies = {
            "a": "b",
            "b": "c",
            "c": "a",  # This creates a cycle
        }

        location_filters = []
        for id_, parent_id in dependencies.items():
            location_filter = LocationFilterWrite(
                external_id=id_,
                name=f"Location {id_}",
                parent_id=-1,  # Deferred lookup
                _parent_external_id=parent_id,
            )
            location_filters.append(location_filter)

        with pytest.raises(ToolkitCycleError) as exc:
            LocationFilterCRUD.topological_sort(location_filters)

        error = exc.value
        assert isinstance(error, ToolkitCycleError)
        assert "cycle" in str(error).lower()
        assert "location filters" in str(error).lower()



from collections.abc import Iterator, Mapping, Sequence
from typing import ClassVar, cast

from cognite.client.data_classes.data_modeling import InstanceApply, NodeId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId
from cognite_toolkit._cdf_tk.constants import MISSING_EXTERNAL_ID, MISSING_INSTANCE_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.storageio import (
    HierarchyIO,
    InstanceIO,
    UploadableStorageIO,
)
from cognite_toolkit._cdf_tk.storageio._base import Page, UploadItem
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, ItemsRequest, SuccessResponseItems
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricKind,
    AssetCentricType,
    JsonVal,
    T_AssetCentricResource,
)

from .data_classes import AssetCentricMapping, AssetCentricMappingList, MigrationMapping, MigrationMappingList
from .data_model import INSTANCE_SOURCE_VIEW_ID
from .selectors import AssetCentricMigrationSelector, MigrateDataSetSelector, MigrationCSVFileSelector


class AssetCentricMigrationIO(
    UploadableStorageIO[AssetCentricMigrationSelector, AssetCentricMapping[T_AssetCentricResource], InstanceApply]
):
    KIND = "AssetCentricMigration"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = InstanceIO.UPLOAD_ENDPOINT

    PENDING_INSTANCE_ID_ENDPOINT_BY_KIND: ClassVar[Mapping[AssetCentricKind, str]] = {
        "TimeSeries": "/timeseries/set-pending-instance-ids",
        "FileMetadata": "/files/set-pending-instance-ids",
    }

    def __init__(self, client: ToolkitClient, skip_linking: bool = True) -> None:
        super().__init__(client)
        self.hierarchy = HierarchyIO(client)
        self.skip_linking = skip_linking

    def as_id(self, item: AssetCentricMapping) -> str:
        return f"{item.mapping.resource_type}_{item.mapping.id}"

    def stream_data(self, selector: AssetCentricMigrationSelector, limit: int | None = None) -> Iterator[Page]:
        if isinstance(selector, MigrationCSVFileSelector):
            iterator = self._stream_from_csv(selector, limit)
        elif isinstance(selector, MigrateDataSetSelector):
            iterator = self._stream_given_dataset(selector, limit)
        else:
            raise ToolkitNotImplementedError(f"Selector {type(selector)} is not supported for stream_data")
        yield from (Page(worker_id="main", items=items) for items in iterator)

    def _stream_from_csv(
        self, selector: MigrationCSVFileSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[T_AssetCentricResource]]]:
        items = selector.items
        if limit is not None:
            items = MigrationMappingList(items[:limit])
        chunk: list[AssetCentricMapping[T_AssetCentricResource]] = []
        for current_batch in chunker_sequence(items, self.CHUNK_SIZE):
            resources = self.hierarchy.get_resource_io(selector.kind).retrieve(current_batch.get_ids())
            for mapping, resource in zip(current_batch, resources, strict=True):
                chunk.append(AssetCentricMapping(mapping=mapping, resource=resource))
            if chunk:
                yield chunk
                chunk = []

    def count(self, selector: AssetCentricMigrationSelector) -> int | None:
        if isinstance(selector, MigrationCSVFileSelector):
            return len(selector.items)
        elif isinstance(selector, MigrateDataSetSelector):
            return self.hierarchy.count(selector.as_asset_centric_selector())
        else:
            raise ToolkitNotImplementedError(f"Selector {type(selector)} is not supported for count")

    def _stream_given_dataset(
        self, selector: MigrateDataSetSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[T_AssetCentricResource]]]:
        asset_centric_selector = selector.as_asset_centric_selector()
        for data_chunk in self.hierarchy.stream_data(asset_centric_selector, limit):
            mapping_list = AssetCentricMappingList[T_AssetCentricResource]([])
            for resource in data_chunk.items:
                # We got the resource from a dataset selector, so we know it is there
                data_set_id = cast(int, resource.data_set_id)
                space_source = self.client.migration.space_source.retrieve(data_set_id=data_set_id)
                instance_space = space_source.instance_space if space_source else None
                if instance_space is None:
                    instance_space = MISSING_INSTANCE_SPACE
                external_id = resource.external_id
                if external_id is None:
                    external_id = MISSING_EXTERNAL_ID.format(project=self.client.config.project, id=resource.id)
                mapping = MigrationMapping(
                    resource_type=self._kind_to_resource_type(selector.kind),
                    instance_id=NodeId(
                        space=instance_space,
                        external_id=external_id,
                    ),
                    id=resource.id,
                    data_set_id=resource.data_set_id,
                    ingestion_view=selector.ingestion_mapping,
                    preferred_consumer_view=selector.preferred_consumer_view,
                )
                mapping_list.append(AssetCentricMapping(mapping=mapping, resource=resource))
            yield mapping_list

    @staticmethod
    def _kind_to_resource_type(kind: AssetCentricKind) -> AssetCentricType:
        mapping: dict[AssetCentricKind, AssetCentricType] = {
            "Assets": "asset",
            "Events": "event",
            "TimeSeries": "timeseries",
            "FileMetadata": "file",
        }
        try:
            return mapping[kind]
        except KeyError as e:
            raise ToolkitNotImplementedError(f"Kind '{kind}' is not supported") from e

    def data_to_json_chunk(
        self,
        data_chunk: Sequence[AssetCentricMapping[T_AssetCentricResource]],
        selector: AssetCentricMigrationSelector | None = None,
    ) -> list[dict[str, JsonVal]]:
        return [item.dump() for item in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceApply:
        raise NotImplementedError()

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[InstanceApply]],
        http_client: HTTPClient,
        selector: AssetCentricMigrationSelector | None = None,
    ) -> Sequence[HTTPMessage]:
        """Upload items by first linking them using files/set-pending-instance-ids and then uploading the instances."""
        if self.skip_linking:
            return list(super().upload_items(data_chunk, http_client, None))
        elif selector is None:
            raise ToolkitNotImplementedError(f"Selector must be provided for uploading {self.KIND} items.")
        elif selector.kind not in self.PENDING_INSTANCE_ID_ENDPOINT_BY_KIND:
            return list(super().upload_items(data_chunk, http_client, None))

        pending_instance_id_endpoint = self.PENDING_INSTANCE_ID_ENDPOINT_BY_KIND[selector.kind]
        results: list[HTTPMessage] = []
        to_upload = self.link_asset_centric(data_chunk, http_client, results, pending_instance_id_endpoint)
        if to_upload:
            results.extend(list(super().upload_items(to_upload, http_client, None)))
        return results

    @classmethod
    def link_asset_centric(
        cls,
        data_chunk: Sequence[UploadItem[InstanceApply]],
        http_client: HTTPClient,
        results: list[HTTPMessage],
        pending_instance_id_endpoint: str,
    ) -> Sequence[UploadItem[InstanceApply]]:
        """Links asset-centric resources to their (uncreated) instances using the pending-instance-ids endpoint."""
        config = http_client.config
        successful_linked: set[str] = set()
        for batch in chunker_sequence(data_chunk, cls.CHUNK_SIZE):
            batch_results = http_client.request_with_retries(
                message=ItemsRequest(
                    endpoint_url=config.create_api_url(pending_instance_id_endpoint),
                    method="POST",
                    api_version="alpha",
                    items=[
                        UploadItem(source_id=item.source_id, item=cls.as_pending_instance_id(item.item))
                        for item in batch
                    ],
                )
            )
            for res in batch_results:
                if isinstance(res, SuccessResponseItems):
                    successful_linked.update(res.ids)
            results.extend(batch_results)
        to_upload = [item for item in data_chunk if item.source_id in successful_linked]
        return to_upload

    @staticmethod
    def as_pending_instance_id(item: InstanceApply) -> PendingInstanceId:
        """Convert an InstanceApply to a PendingInstanceId for linking."""
        source = next((source for source in item.sources if source.source == INSTANCE_SOURCE_VIEW_ID), None)
        if source is None:
            raise ValueError(f"Cannot extract ID from item of type {type(item).__name__!r}")
        if not isinstance(source.properties["id"], int):
            raise ValueError(f"Unexpected ID type: {type(source.properties['id']).__name__!r}")
        id_ = source.properties["id"]
        return PendingInstanceId(
            pending_instance_id=NodeId(item.space, item.external_id),
            id=id_,
        )



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.robotics.data_postprocessing import RobotDataPostProcessingYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def valid_robot_data_postprocessing_cases() -> Iterable:
    yield pytest.param(
        {
            "name": "Read dial gauge",
            "externalId": "read_dial_gauge",
            "method": "read_dial_gauge",
            "description": "Original Description",
            "inputSchema": {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "id": "robotics/schemas/0.1.0/capabilities/ptz",
                "title": "PTZ camera capability input",
                "type": "object",
                "properties": {
                    "method": {"type": "string"},
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "tilt": {"type": "number", "minimum": -90, "maximum": 90},
                            "pan": {"type": "number", "minimum": -180, "maximum": 180},
                            "zoom": {"type": "number", "minimum": 0, "maximum": 100},
                        },
                        "required": ["tilt", "pan", "zoom"],
                    },
                },
                "required": ["method", "parameters"],
                "additionalProperties": False,
            },
        },
        id="full-data-postprocessing",
    )
    yield pytest.param(
        {"name": "Defect Detection", "externalId": "defect_detection_001", "method": "detect_defects"},
        id="minimal-data-postprocessing",
    )


def invalid_robot_data_postprocessing_cases() -> Iterable:
    yield pytest.param(
        {"name": "Missing External ID", "method": "some_method"},
        {"Missing required field: 'externalId'"},
        id="missing-required-field",
    )
    yield pytest.param(
        {"name": "Test PostProcessing", "externalId": "postproc_001", "method": "test_method", "unknownField": "value"},
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )


class TestRobotDataPostProcessingYAML:
    @pytest.mark.parametrize("data", list(find_resources("DataPostProcessing", "robotics")))
    def test_load_valid_robot_data_postprocessing_file(self, data: dict[str, object]) -> None:
        loaded = RobotDataPostProcessingYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", valid_robot_data_postprocessing_cases())
    def test_load_valid_robot_data_postprocessing(self, data: dict[str, object]) -> None:
        loaded = RobotDataPostProcessingYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", invalid_robot_data_postprocessing_cases())
    def test_invalid_robot_data_postprocessing(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, RobotDataPostProcessingYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors
